import sympy as sp  # type: ignore
from numpy import cos, sin

import autokoopman.core.system as asys


class Spring(asys.SymbolicContinuousSystem):
    r"""
    Spring pendulum
        We study the behavior of the planar spring-pendulum described in [1].
        It consists of a solid ball of mass :math:`m` and a spring of natural length :math:`L`.
        The spring constant is :math:`k`.

        We study the evolutions of the length r of the spring and the angle :math:`\theta`
        between the spring and the vertical. They are modeled by the following differential
        equations.

        The constants are set as :math:`k=2`, :math:`L=1`, and :math:`g=9.8`.

        .. math::
            \begin{bmatrix} \dot r \\ \dot \theta \\ \dot d_r \\ \dot d_{\theta} \end{bmatrix} = \begin{bmatrix} d_r \\ d_{\theta} \\ r d_{\theta}^2 + g * \cos (\theta) -2 (r - 1) \\ \left( -2 d_r d_{\theta} - g \sin (\theta) \right) / 2 \end{bmatrix}

        Initial set:
            r in [1.19,1.21]
            theta in [0.49,0.51]
            dr in [0,0]
            dtheta in [0,0]

    References
        J. D. Meiss. Differential Dynamical Systems (Monographs on Mathematical Modeling and Computation), Book 14, SIAM publishers, 2007.
    """

    def __init__(self, g=9.81):
        r, theta, dr, dtheta = sp.symbols("r theta dr dtheta")
        xdot = [
            dr,
            dtheta,
            r * dtheta**2 + g * cos(theta) - 2 * (r - 1),
            (-2 * dr * dtheta - g * sin(theta)) / r,
        ]
        super(Spring, self).__init__((r, theta, dr, dtheta), xdot)



import torch
import torch.nn as nn
import numpy as np
from typing import Optional, Tuple
import tqdm
import autokoopman.core.system as ksys
import autokoopman.core.estimator as kest
import autokoopman.core.trajectory as ktraj


class DeepKoopman(kest.TrajectoryEstimator):
    r"""
    Deep Learning Koopman with Inputs

    This is a simple implementation of an AutoEncoder architecture to learn the Koopman observables and
    operator. It allows an input space, which is directly injected into the observables space (like KIC).
    To improve a good fit, a *metric loss* is introduced to make the distances between states approximately
    preserve in the observables space.

    The loss is defined as

    .. math::
        \mathcal L(\mathbf x, \mathbf x_r,\mathbf x', \mathbf x'_r, \mathbf y, \mathbf y') = \left\| \mathbf x-\mathbf x_r \right\|_2^2 + \lambda_p \left\| \mathbf x'-\mathbf x'_r \right\|_2^2 + \lambda_m \left| \left\| \mathbf y' - \mathbf y \right\|_2^2 -  \left\| \mathbf x' - \mathbf x \right\|_2^2 \right|

    where :math:`\mathbf x` is the current state, :math:`\mathbf x_r` is the reconstructed current state,
    :math:`\mathbf x'` is the next state, :math:`\mathbf x'_r` is the reconstructed state,
    :math:`\mathbf y` is the current observable, and :math:`\mathbf y'` is the next observable.

    :param state_dim: system state space dimension
    :param input_dim: system input space dimension
    :param hidden_dim: observables space dimension
    :param pred_loss_weight: weighting factor for prediction loss (of total loss)
    :param metric_loss_weight: weighting factor for metric loss (of total loss)
    :param max_loss_weight: weighting factor against max loss (0 if outliers)
    :param weight_decay_weight: weight decay
    :param hidden_enc_dim: dimension of hidden layers in the encoder/decoder
    :param max_iter: maximum number of iterations
    :param lr: learning rate
    :param validation_data: additional uniform time trajectories to validate at each iteration
    :param rollout_steps: number of rollout steps
    :param encoder_module: PyTorch Module Encoder (pass in externally)
    :param decoder_module: PyTorch Module Decoder (pass in externally)
    :param torch_device: device to run PyTorch (if None, it will attempt to use cuda:0)
    :param display_progress: show progress bar
    :param verbose: set verbosity

    References
        Li, Y., He, H., Wu, J., Katabi, D., & Torralba, A. (2019).
        Learning compositional koopman operators for model-based control.
        arXiv preprint arXiv:1910.08264.
    """

    def __init__(
        self,
        state_dim: int,
        input_dim: int,
        hidden_dim: int,
        pred_loss_weight: float = 0.5,
        metric_loss_weight: float = 1e-5,
        max_loss_weight: float = 1e-5,
        weight_decay_weight: float = 1e-7,
        hidden_enc_dim: int = 32,
        max_iter: int = 500,
        lr: float = 1e-3,
        validation_data: Optional[ktraj.UniformTimeTrajectoriesData] = None,
        num_hidden_layers: int = 1,
        rollout_steps: int = 4,
        encoder_module: Optional[nn.Module] = None,
        decoder_module: Optional[nn.Module] = None,
        torch_device: Optional[str] = None,
        display_progress: bool = True,
        verbose: bool = True,
    ):
        hidden_enc_dim = int(hidden_enc_dim)
        num_hidden_layers = int(num_hidden_layers)
        hidden_dim = int(hidden_dim)

        self.names = None
        self.max_iter = max_iter
        self.lr = lr
        self.validation_data = validation_data
        self.verbose = verbose
        self.disp_progress = display_progress

        self.has_input = input_dim > 0

        if torch_device is None:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            if self.verbose:
                print(f"DeepKoopman is using torch device '{self.device}'")
        else:
            self.device = torch_device

        self.hidden_dim = hidden_dim

        hidden_layers = [
            nn.Linear(hidden_enc_dim, hidden_enc_dim),
            nn.PReLU(),
        ] * num_hidden_layers

        if encoder_module is None:
            self.encoder = nn.Sequential(
                nn.Linear(state_dim, hidden_enc_dim),
                nn.PReLU(),
                *(hidden_layers),
                nn.Linear(hidden_enc_dim, hidden_dim),
                nn.Tanh(),
            )
        else:
            self.encoder = encoder_module

        if decoder_module is None:
            self.decoder = nn.Sequential(
                nn.Linear(hidden_dim, hidden_enc_dim),
                nn.PReLU(),
                *(
                    [nn.Linear(hidden_enc_dim, hidden_enc_dim), nn.PReLU()]
                    * num_hidden_layers
                ),
                nn.Linear(hidden_enc_dim, state_dim),
                nn.Tanh(),
            )
        else:
            self.decoder = decoder_module

        self.propagate = nn.Linear(hidden_dim + input_dim, hidden_dim, bias=False)

        # move the modules to the target device
        self.encoder.to(self.device)
        self.decoder.to(self.device)
        self.propagate.to(self.device)

        # TODO
        self.loss_hist = None
        self.metric_loss_weight = metric_loss_weight
        self.pred_loss_weight = pred_loss_weight
        self.max_loss_weight = max_loss_weight
        self.weight_decay_weight = weight_decay_weight
        self.rollout_steps = rollout_steps

    def forward(self, x, u, step_size):
        """
        NN Forward Function

        :param x: current system state
        :param u: current system input
        """
        # get current step observable
        y = self.encoder(x)

        # get next step observable
        if u is not None:
            yn = self.propagate(torch.cat((y, u), axis=-1))
        else:
            yn = self.propagate(y)
        if step_size > 1:
            for _ in range(step_size - 1):
                if u is not None:
                    yn = self.propagate(torch.cat((yn, u), axis=-1))
                else:
                    yn = self.propagate(yn)

        # get next step state
        xn = self.decoder(yn)

        # get x reconstruction
        xr = self.decoder(y)

        return x, xn, y, yn, xr

    @property
    def system_matrices(self) -> Tuple[np.ndarray, np.ndarray]:
        """get the linear system matrices :math:`\mathbf x_{n+1} = A \mathbf x_n + B \mathbf u_n`

        :returns: A, B
        """
        weight = self.propagate.weight.data.numpy()
        A = weight[:, : self.hidden_dim]
        B = weight[:, self.hidden_dim :]
        return A, B

    @property
    def model(self) -> ksys.System:
        """
        packs the autoencoder into a system object
        """

        def step_func(t, x, i):
            with torch.no_grad():
                tx = torch.tensor(x, dtype=torch.float32).to(self.device)
                if self.has_input:
                    tu = torch.tensor(i, dtype=torch.float32).to(self.device)
                    tu /= self.u_mult
                else:
                    tu = None
                tx = (tx - self.xmean) / self.xstd
                _, xn, _, _, _ = self.forward(tx, tu, 1)
                xn = xn * self.xstd + self.xmean
                return xn.cpu().detach().numpy()

        return ksys.StepDiscreteSystem(step_func, self.names)

    def fit(self, X: ktraj.TrajectoriesData) -> None:
        """fits the discrete system model

        :param X: trajectories
        """
        self.train(X)

    def train(self, trajs):
        """
        Train the AutoEncoder

        Internally updates the loss_hist field to record training information.

        TODO: implement early stopping

        :param X: snapshot of states
        :param Xn: snapshot of next states
        :param U: snapshot of inputs corresponding to X
        """
        assert isinstance(
            trajs, ktraj.UniformTimeTrajectoriesData
        ), "trajs must be uniform time"
        self.sampling_period = trajs.sampling_period
        self.names = trajs.state_names

        mseloss = nn.MSELoss()
        l1loss = nn.L1Loss()

        mselossv = nn.MSELoss()
        l1lossv = nn.L1Loss()

        # upload single step matrices to the compute device
        X, Xn, U = trajs.next_step_matrices
        X = torch.tensor(X.T, dtype=torch.float32).to(self.device)
        Xn = torch.tensor(Xn.T, dtype=torch.float32).to(self.device)
        U = (
            torch.tensor(U.T, dtype=torch.float32).to(self.device)
            if self.has_input
            else None
        )

        # get the scaling factors
        self.x_mult = torch.max(torch.abs(X))
        self.xmean = torch.mean(X, dim=0)
        self.xstd = self.x_mult  # torch.std(X, dim=0)

        # normalize it
        nX = (X - self.xmean) / self.xstd
        nXn = (Xn - self.xmean) / self.xstd

        if self.has_input:
            self.u_mult = torch.max(torch.abs(U))
            nU = U / self.u_mult
        else:
            nU = None

        # upload validation data to device if needed
        if self.validation_data is not None:
            # transpose necessary to change construction
            _Xv, _Xnv, _Uv = self.validation_data.next_step_matrices
            Xv = torch.tensor(_Xv.T, dtype=torch.float32).to(self.device)
            Xnv = torch.tensor(_Xnv.T, dtype=torch.float32).to(self.device)
            nXv = (Xv - self.xmean) / self.xstd
            nXnv = (Xnv - self.xmean) / self.xstd
            if self.has_input:
                Uv = torch.tensor(_Uv.T, dtype=torch.float32).to(self.device)
                nUv = Uv / self.u_mult
            else:
                nUv = None
        else:
            nXv = None
            nUv = None

        def _get_loss(mseloss, l1loss, x, xn, y, yn, xr, Xn):
            # reconstruction and prediction loss
            ae_loss = mseloss(x, xr)
            pred_loss = mseloss(xn, Xn)

            # linearity loss
            lin_loss = mseloss(yn, self.encoder(Xn))

            # largest loss
            inf_loss = torch.max(torch.abs(x - xr)) + torch.max(torch.abs(Xn - xn))

            # metric loss
            metric_loss = l1loss(torch.norm(yn - y, dim=1), torch.norm(xn - x, dim=1))

            # frobenius norm of operator
            weight_loss = 0
            for l in self.encoder:
                if isinstance(l, nn.Linear):
                    weight_loss += torch.norm(l.weight.data)
            for l in self.decoder:
                if isinstance(l, nn.Linear):
                    weight_loss += torch.norm(l.weight.data)

            total_loss = (
                self.pred_loss_weight * (ae_loss + pred_loss)
                + lin_loss
                + self.max_loss_weight * inf_loss
                + self.weight_decay_weight * weight_loss
                + self.metric_loss_weight * metric_loss
            )
            return {
                "total_loss": total_loss,
                "recon_loss": ae_loss,
                "pred_loss": pred_loss,
                "lin_loss": lin_loss,
                "inf_loss": inf_loss,
                "weight_loss": weight_loss,
                "metric_loss": metric_loss,
            }

        encoder_optimizer = torch.optim.Adam(self.encoder.parameters(), lr=self.lr)
        decoder_optimizer = torch.optim.Adam(self.decoder.parameters(), lr=self.lr)
        propagate_optimizer = torch.optim.Adam(self.propagate.parameters(), lr=self.lr)

        loss_hist = {}

        for _ in (
            tqdm.tqdm(range(self.max_iter))
            if (self.verbose and self.disp_progress)
            else range(self.max_iter)
        ):
            risk = None
            # roll out over several steps
            for step_size in range(1, self.rollout_steps + 1):
                # upload to GPU and normalize the data
                # NOTE: this is expensive :(
                X, Xn, U = trajs.n_step_matrices(step_size)
                X = torch.tensor(X.T, dtype=torch.float32).to(self.device)
                Xn = torch.tensor(Xn.T, dtype=torch.float32).to(self.device)
                U = (
                    torch.tensor(U.T, dtype=torch.float32).to(self.device)
                    if self.has_input
                    else None
                )
                nX = (X - self.xmean) / self.xstd
                nXn = (Xn - self.xmean) / self.xstd
                if self.has_input:
                    nU = U / self.u_mult
                else:
                    nU = None

                # get risk (loss)
                # sum over all rollouts
                x, xn, y, yn, xr = self.forward(nX, nU, step_size)
                if risk is None:
                    risk = _get_loss(mseloss, l1loss, x, xn, y, yn, xr, nXn)
                else:
                    _risk = _get_loss(mseloss, l1loss, x, xn, y, yn, xr, nXn)
                    for k, v in _risk.items():
                        risk[k] += v

            # store total loss
            for k, v in risk.items():
                if k in loss_hist:
                    loss_hist[k].append(v.cpu().detach().numpy())
                else:
                    loss_hist[k] = [v.cpu().detach().numpy()]

            # optimize
            encoder_optimizer.zero_grad()
            decoder_optimizer.zero_grad()
            propagate_optimizer.zero_grad()

            risk["total_loss"].backward()

            encoder_optimizer.step()
            decoder_optimizer.step()
            propagate_optimizer.step()

            # do validation if applicable
            if self.validation_data is not None:
                with torch.no_grad():
                    x, xn, y, yn, xr = self.forward(nXv, nUv, 1)
                    loss = _get_loss(mselossv, l1lossv, x, xn, y, yn, xr, nXnv)

                    # add to the loss history with validation prefix
                    for k, v in loss.items():
                        k = f"validation_{k}"
                        if k in loss_hist:
                            loss_hist[k].append(v.cpu().detach().numpy())
                        else:
                            loss_hist[k] = [v.cpu().detach().numpy()]

        self.loss_hist = loss_hist



import itertools
from autokoopman.autokoopman import (
    obs_types,
    opt_types,
    scoring_func_types,
    auto_koopman,
)
from autokoopman.core.trajectory import UniformTimeTrajectoriesData
import numpy as np
import pytest


auto_config = tuple(
    itertools.product(obs_types, opt_types, scoring_func_types, {True, False})
)


@pytest.mark.parametrize(
    "obs, opt, cost, normalize",
    auto_config,
)
def test_autokoopman(obs, opt, cost, normalize):
    # for a complete example, let's create an example dataset using an included benchmark system
    import autokoopman.benchmark.fhn as fhn

    fhn = fhn.FitzHughNagumo()
    np.random.seed(0)

    # given issue #29, let's make these differently sized
    ivs = np.random.uniform(low=-2.0, high=2.0, size=(20, 2))
    training_data = UniformTimeTrajectoriesData(
        {
            idx: fhn.solve_ivp(
                initial_state=iv,
                tspan=[0.0, np.random.random() + 1.0],
                sampling_period=0.1,
            )
            for idx, iv in enumerate(ivs)
        }
    )

    # learn model from data
    # make the run as short as possible but still be meaningful for catching errors
    experiment_results = auto_koopman(
        training_data,
        sampling_period=0.1,
        obs_type=obs,
        opt=opt,
        cost_func=cost,
        n_obs=20,
        max_opt_iter=2,
        grid_param_slices=2,
        n_splits=2,
        rank=(10, 12, 1),
        max_epochs=1,
        torch_device="cpu",
        normalize=normalize,
    )



import autokoopman.estimator.sindy as asindy
import autokoopman.benchmark.fhn as afhn
import numpy as np
import pytest


@pytest.mark.skip(reason="pySINDy broke on some machines")
def test_sindy():
    fhn = afhn.FitzHughNagumo()

    training_ivs = np.random.random((10, 2))
    training_data = fhn.solve_ivps(training_ivs, (0.0, 3.0), sampling_period=0.05)

    sindy = asindy.PolynomialSindy(["x0", "x1"], (0.05, 2.0))
    sindy_model = sindy.get_model([0.1, 2])
    sindy_model.fit(training_data)



def test_deepk():
    """test that deep koopman runs"""
    # the notebook imports
    import numpy as np
    import autokoopman.benchmark.fhn as fhn

    # make a training and validation set
    dt = 0.1
    fhn = fhn.FitzHughNagumo()
    training_data = fhn.solve_ivps(
        initial_states=np.random.uniform(low=-4.0, high=4.0, size=(100, 2)),
        tspan=[0.0, 3.0],
        sampling_period=dt,
    )

    test_ivs = np.random.uniform(low=-2.0, high=2.0, size=(10, 2))
    test_data = fhn.solve_ivps(
        initial_states=test_ivs, tspan=[0.0, 10.0], sampling_period=dt
    )

    # run deepk
    import autokoopman.estimator.deepkoopman as dk

    koop = dk.DeepKoopman(
        2,
        0,
        40,
        max_iter=10,  # very low count for bring up test
        lr=1e-3,
        hidden_enc_dim=64,
        num_hidden_layers=2,
        metric_loss_weight=0.3,
        pred_loss_weight=20.0,
        torch_device="cpu",  #  don't require a GPU!
        validation_data=test_data,
    )
    koop.fit(training_data)



import abc
import random
from typing import Sequence, Callable, TypedDict, Any
import numpy as np

from autokoopman.core.estimator import TrajectoryEstimator
from autokoopman.core.trajectory import (
    TrajectoriesData,
    UniformTimeTrajectory,
    UniformTimeTrajectoriesData,
)
from autokoopman.core.format import _clip_list
from sklearn.model_selection import KFold


class Parameter:
    def __init__(self, name):
        self._name = name

    @abc.abstractmethod
    def random(self):
        pass

    @abc.abstractmethod
    def is_member(self, item) -> bool:
        ...

    def __contains__(self, item) -> bool:
        return self.is_member(item)

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return f"<{self.__class__.__name__} Name: {self.name}>"


class FiniteParameter(Parameter):
    def __init__(self, name: str, elements: Sequence):
        super(FiniteParameter, self).__init__(name)
        self.elements = tuple(elements)

    def is_member(self, item) -> bool:
        return item in self.elements

    def random(self):
        return random.choice(self.elements)


class ContinuousParameter(Parameter):
    @staticmethod
    def loguniform(low=0.1, high=1, size=None):
        return np.exp(np.random.uniform(np.log(low), np.log(high), size))

    @staticmethod
    def uniform(low=0, high=1, size=None):
        return np.random.uniform(low, high, size)

    def __init__(self, name: str, domain_lower, domain_upper, distribution="uniform"):
        super(ContinuousParameter, self).__init__(name)
        assert domain_upper >= domain_lower
        self._interval = (domain_lower, domain_upper)
        self.distribution = distribution

    def is_member(self, item) -> bool:
        return item >= self._interval[0] and item <= self._interval[1]

    def random(self):
        if isinstance(self.distribution, Callable):
            return self.distribution()
        elif hasattr(self, self.distribution):
            return getattr(self, self.distribution)(
                self._interval[0], self._interval[1]
            )
        else:
            raise ValueError(f"cannot find distribution {self.distribution}")


class DiscreteParameter(FiniteParameter):
    def __init__(self, name: str, domain_lower: int, domain_upper: int, step=1):
        super(DiscreteParameter, self).__init__(
            name, range(domain_lower, domain_upper, step)
        )


class ParameterSpace(Parameter):
    def __init__(self, name: str, coords: Sequence[Parameter]):
        super(ParameterSpace, self).__init__(name)
        self._coords = coords
        self._cdict = {c.name: c for c in self._coords}

    def is_member(self, item) -> bool:
        return all([itemi in coordi for itemi, coordi in zip(item, self._coords)])

    def random(self):
        return [coordi.random() for coordi in self._coords]

    def __getitem__(self, item):
        assert (
            item in self._cdict
        ), f"coordinate {item} was not found in space (values are {list(self._cdict.keys())})"
        return self._cdict[item]

    def __iter__(self):
        for c in self._coords:
            yield c

    @property
    def dimension(self):
        return len(self._coords)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} Name: {self.name} Dimensions: {self.dimension} "
            f"Coordinates: {_clip_list([s.name+f': {s.__class__.__name__}' for s in self._coords])}>"
        )


class HyperparameterMap:
    """
    define and associate a hyperparameter space with a moddel
    """

    def __init__(self, parameter_space: ParameterSpace):
        self.parameter_space = parameter_space

    @abc.abstractmethod
    def get_model(self, hyperparams: Sequence) -> TrajectoryEstimator:
        raise NotImplementedError


class TuneResults(TypedDict):
    model: TrajectoryEstimator
    param: Sequence[Any]
    score: float


class TrajectoryScoring:
    @staticmethod
    def end_point_score(true_data: TrajectoriesData, prediction_data: TrajectoriesData):
        errors = (prediction_data - true_data).norm()
        end_errors = np.array([s.states[-1] for s in errors])
        return np.mean(end_errors)

    @staticmethod
    def total_score(true_data: TrajectoriesData, prediction_data: TrajectoriesData):
        errors = (prediction_data - true_data).norm()
        end_errors = np.array([s.states.flatten() for s in errors])

        return np.mean(np.concatenate(end_errors, axis=0))

    @staticmethod
    def relative_score(true_data: TrajectoriesData, prediction_data: TrajectoriesData):
        # TODO: implement this
        err_term = []
        for k in prediction_data.traj_names:
            pred_t = prediction_data[k]
            true_t = true_data[k]
            abs_error = np.linalg.norm(pred_t.states - true_t.states)
            mean_error = np.linalg.norm(pred_t.states - np.mean(pred_t.states, axis=0))
            err_term.append(abs_error / mean_error)
        return np.mean(np.array(err_term))


class HyperparameterTuner(abc.ABC):
    """
    Tune a HyperparameterMap Object
        The tuner implements a model learning pipeline that can take a mapping from hyperparameters to a model and
        training data and learn an optimized model. For example, given a SINDy estimator, hyperparameter tuner can
        implement gridsearch to find the best sparsity threshold and library type.
    """

    @staticmethod
    def generate_predictions(
        trained_model: TrajectoryEstimator, holdout_data: TrajectoriesData
    ):
        preds = {}
        # get the predictions
        for k, v in holdout_data._trajs.items():
            # ugh, this is a hack--sampling period is meaningful if the system is discrete
            # TODO: make the behavior of sampling period less problematic
            _kwargs = {}
            if hasattr(v, "sampling_period"):
                _kwargs["sampling_period"] = v.sampling_period
            sivp_interp = trained_model.model.solve_ivp(
                v.states[0],
                (np.min(v.times), np.max(v.times)),
                inputs=v.inputs,
                teval=v.times,
                **_kwargs,
            )
            preds[k] = sivp_interp
        return TrajectoriesData(preds)

    def __init__(
        self,
        parameter_model: HyperparameterMap,
        training_data: TrajectoriesData,
        n_splits=None,
        display_progress: bool = True,
        verbose: bool = True,
    ):
        self._parameter_model = parameter_model
        self._training_data = training_data
        self.scores = []
        self.best_scores = []
        self.best_result = None
        self.error_messages = []
        self.n_splits = n_splits
        self.verbose = verbose
        self.disp_progress = display_progress

    def _reset_scores(self):
        self.scores = []
        self.best_scores = []
        self.best_result = None

    @abc.abstractmethod
    def tune(
        self,
        nattempts=100,
        scoring_func: Callable[
            [TrajectoriesData, TrajectoriesData], float
        ] = TrajectoryScoring.total_score,
    ) -> TuneResults:
        pass

    def tune_sampling(
        self,
        nattempts,
        scoring_func: Callable[[TrajectoriesData, TrajectoriesData], float],
    ):
        import tqdm

        self._reset_scores()
        best_model = None
        best_params = None
        best_score = None
        score = None

        if self.n_splits is not None:
            kf = KFold(n_splits=self.n_splits)

        for _ in (
            tqdm.tqdm(
                range(nattempts),
                total=nattempts,
                desc=f"Tuning {self.__class__.__name__}",
            )
            if (self.verbose and self.disp_progress)
            else range(nattempts)
        ):
            param = yield

            assert isinstance(param, Sequence), "yielded param must be a sequence"
            model = self._parameter_model.get_model(param)

            # have something, even if it can't be optimized
            if len(self.scores) == 0:
                self.best_result = TuneResults(
                    model=model,
                    param=param,
                    score=np.infty,
                )

            if self.n_splits is None:
                model.fit(self._training_data)
                prediction_data = self.generate_predictions(model, self._training_data)
                score = scoring_func(self._training_data, prediction_data)
            else:
                iscores = []
                names = np.array(list(self._training_data.traj_names))[:, np.newaxis]
                for train_index, test_index in kf.split(names):
                    # if all trajectories are uniform time, make the data uniform time
                    if all(
                        [
                            isinstance(tinst, UniformTimeTrajectory)
                            for tinst in self._training_data
                        ]
                    ):
                        TrajsType = UniformTimeTrajectoriesData
                    else:
                        TrajsType = TrajectoriesData

                    # create test train split
                    training = TrajsType(
                        {ti[0]: self._training_data[ti[0]] for ti in names[train_index]}
                    )
                    validation = TrajsType(
                        {ti[0]: self._training_data[ti[0]] for ti in names[test_index]}
                    )

                    # do the fit and scoring (select the best median)
                    model = self._parameter_model.get_model(param)
                    model.fit(training)
                    prediction_data = self.generate_predictions(model, validation)
                    iscore = scoring_func(validation, prediction_data)
                    iscores.append(iscore)
                score = np.median(iscores)

            if len(self.scores) == 0 or score < np.min(self.scores):
                best_model = model
                best_params = param
                best_score = score
                assert (
                    (best_model is not None)
                    and (best_params is not None)
                    and (best_score is not None)
                )
                self.best_result = TuneResults(
                    model=best_model, param=best_params, score=best_score
                )
            self.scores.append(score if not np.isnan(score) else 1e12)
            self.best_scores.append(np.min(self.scores))
            yield score



import sympy as sp  # type: ignore

import autokoopman.core.system as asys


class ProdDestr(asys.SymbolicContinuousSystem):
    r"""
    Production Destruction System (PRDE20)
        In this model, :math:`x_0` is the nutrients, :math:`x_1` the phytoplankton and :math:`x_2` the detritus.
        The constraints are :math:`x_0(t), x_1(t), x_2(t)` are positive and :math:`x_0(t) + x_1(t) + x_2(t) = 10` for all t.

        The initial condition :math:`x_0(0) = 9.98`, :math:`x_1(0) = 0.01` and :math:`x_2(0) = 0.01`

        :math:`I: x_0(0) \in [9.5, 10.0]`, i.e., uncertainty on the initial condition;

        From Michaelis-Menten Theory, the evolution function is

        .. math::
            \begin{bmatrix} \dot x_0 \\ \dot x_1 \\ \dot x_2 \end{bmatrix} = \begin{bmatrix} \frac{-x_0 x_1}{1 + x_0} \\ \frac{x_0 x_1}{1 + x_0} - 0.3 x_1 \\ 0.3 x_1 \end{bmatrix}

    :param alpha: model parameter

    Example:
        .. code-block:: python

            import autokoopman.benchmark.prde20 as prde

            sys = prde.ProdDestr()
            traj = sys.solve_ivp(
                initial_state=[9.98, 0.01, 0.01],
                tspan=(0.0, 20.0),
                sampling_period=0.01
            )
            traj.states
            # array([[9.98000000e+00, 1.00000000e-02, 1.00000000e-02],
            #        [9.97990883e+00, 1.00610783e-02, 1.00300915e-02],
            #        [9.97981710e+00, 1.01225295e-02, 1.00603668e-02],
            #        ...,
            #        [3.34608394e-08, 4.41976926e-01, 9.55802304e+00],
            #        [3.33135031e-08, 4.40652987e-01, 9.55934698e+00],
            #        [3.31672496e-08, 4.39333011e-01, 9.56066696e+00]])

    References
        Geretti, L., Sandretto, J. A. D., Althoff, M., Benet, L., Chapoutot, A., Chen, X., ... & Schilling, C. (2020).
        ARCH-COMP20 category report: Continuous and hybrid systems with nonlinear dynamics. EPiC Series in Computing, 74, 49-75. pp 53
    """

    def __init__(self, alpha=0.3):
        x0, x1, x2 = sp.symbols("x0 x1 x2")
        xdot = [(-x0 * x1) / (1 + x0), (x0 * x1) / (1 + x0) - alpha * x1, alpha * x1]
        super(ProdDestr, self).__init__((x0, x1, x2), xdot)



import numpy as np

from autokoopman.core.trajectory import TrajectoriesData
from autokoopman.core.tuner import (
    TuneResults,
    TrajectoryScoring,
    HyperparameterMap,
    ParameterSpace,
    FiniteParameter,
    DiscreteParameter,
    ContinuousParameter,
)
from autokoopman.core.format import hide_prints
import autokoopman.core.tuner as atuner
from typing import Callable


class BayesianOptTuner(atuner.HyperparameterTuner):
    r"""Bayesian Optimization Hyperparameter Tuner"""

    @staticmethod
    def make_bounds(param_space: ParameterSpace):
        """make GPyOpt Optimization Problem domain"""
        bounds = []
        lengths = []
        is_log = []
        for idx, coord in enumerate(param_space):
            if isinstance(coord, ContinuousParameter):
                g_log = coord.distribution == "loguniform"
                g_var = {
                    "name": f"var_{coord.name}_{idx}",
                    "type": "continuous"
                    if isinstance(coord, ContinuousParameter)
                    else "discrete",
                    "domain": (
                        np.log(coord._interval[0]) if g_log else coord._interval[0],
                        np.log(coord._interval[1]) if g_log else coord._interval[1],
                    ),
                }
                g_len = coord._interval[1] - coord._interval[0]
            elif isinstance(coord, DiscreteParameter):
                g_var = {
                    "name": f"var_{coord.name}_{idx}",
                    "type": "continuous"
                    if isinstance(coord, ContinuousParameter)
                    else "discrete",
                    "domain": tuple(coord.elements),
                }
                g_len = max(coord.elements) - min(coord.elements)
                g_log = False
            elif isinstance(coord, FiniteParameter):
                g_var = {
                    "name": f"var_{coord.name}_{idx}",
                    "type": "continuous"
                    if isinstance(coord, ContinuousParameter)
                    else "discrete",
                    "domain": tuple(coord.elements),
                }
                g_len = max(coord.elements) - min(coord.elements)
                g_log = False
            else:
                g_var = None
                g_len = None
                g_log = False
            if g_var is not None:
                bounds.append(g_var)
                lengths.append(g_len)
                is_log.append(g_log)
        return bounds, lengths, is_log

    def __init__(
        self,
        parameter_model: HyperparameterMap,
        training_data: TrajectoriesData,
        n_samps=10,
        **kwargs,
    ):
        super(BayesianOptTuner, self).__init__(parameter_model, training_data, **kwargs)
        self.n_samps = n_samps

    def tune(
        self,
        nattempts=100,
        scoring_func: Callable[
            [TrajectoriesData, TrajectoriesData], float
        ] = TrajectoryScoring.end_point_score,
    ) -> TuneResults:
        import GPyOpt
        import GPy
        import os, sys

        # TODO: switch to something more modern (not GPyOpt?)
        # get GPyOpt domain
        bounds, lengthscales, is_logs = self.make_bounds(
            self._parameter_model.parameter_space
        )

        # create the sampler and send it the parameters
        sampling = self.tune_sampling(nattempts, scoring_func)
        next(sampling)

        def gpy_obj(param):
            """function for the optimizer
            TODO: deal with logspace correctly
            """
            try:
                param_v = tuple(param.flatten())
                param = [
                    (p if not il else np.exp(p)) for il, p in zip(is_logs, param_v)
                ]
                val = sampling.send(param)
                next(sampling)
                return val
            except StopIteration:
                return 1e10
            except Exception as exc:
                print(f"Error:<{exc}>")
                self.error_messages.append((param, exc))
                return 1e10

        kern = GPy.kern.Matern52(
            len(bounds),
            variance=1.0,
            lengthscale=np.array(lengthscales) / 5.0,
            ARD=True,
        )
        with hide_prints():
            self.model = GPyOpt.models.gpmodel.GPModel(kernel=kern, verbose=False)
            self.bopt = GPyOpt.methods.BayesianOptimization(
                gpy_obj, domain=bounds, model=self.model, verbose=False
            )
        self.bopt.run_optimization(max_iter=nattempts)
        return self.best_result



from typing import Optional, Tuple, Sequence

import numpy as np
import pysindy as ps  # type: ignore

import autokoopman.core.estimator as aest
import autokoopman.core.system as asys
import autokoopman.core.trajectory as atraj
import autokoopman.core.tuner as atune


class FourierSindy(atune.HyperparameterMap):
    """SINDy with hyperparameters for Fourier library"""

    def __init__(self, names: Sequence[str], threshold_range: Tuple[float, float]):
        self.names = names
        super(FourierSindy, self).__init__(
            atune.ParameterSpace(
                "sindy-fourier",
                [
                    atune.ContinuousParameter(
                        "sr3-threshold", *threshold_range, distribution="loguniform"
                    ),
                    atune.DiscreteParameter("n-frequency", 1, 4),
                ],
            )
        )

    def get_model(self, hyperparams: Sequence) -> aest.TrajectoryEstimator:
        model = ps.SINDy(
            feature_names=self.names,
            differentiation_method=ps.FiniteDifference(),
            optimizer=ps.SR3(threshold=hyperparams[0], thresholder="l1"),
            feature_library=ps.FourierLibrary(n_frequencies=hyperparams[1]),
        )
        return SindyEstimator(model)


class PolynomialSindy(atune.HyperparameterMap):
    """SINDy with hyperparameters for polynomial library"""

    def __init__(self, names: Sequence[str], threshold_range: Tuple[float, float]):
        self.names = names
        super(PolynomialSindy, self).__init__(
            atune.ParameterSpace(
                "sindy-poly",
                [
                    atune.ContinuousParameter(
                        "sr3-threshold", *threshold_range, distribution="loguniform"
                    ),
                    atune.DiscreteParameter("poly-degree", 1, 4),
                ],
            )
        )

    def get_model(self, hyperparams: Sequence) -> aest.TrajectoryEstimator:
        model = ps.SINDy(
            feature_names=self.names,
            differentiation_method=ps.FiniteDifference(),
            optimizer=ps.SR3(threshold=hyperparams[0], thresholder="l1"),
            feature_library=ps.PolynomialLibrary(degree=hyperparams[1]),
        )
        return SindyEstimator(model)


class SindyEstimator(aest.TrajectoryEstimator):
    def __init__(self, sindy: Optional[ps.SINDy] = None):
        self._usr_sindy = sindy

    def fit(self, X: atraj.TrajectoriesData) -> None:
        if self._usr_sindy is None:
            self._model = ps.SINDy(
                feature_names=X.state_names,
                differentiation_method=ps.FiniteDifference(),
                optimizer=ps.SR3(threshold=0.04, thresholder="l1"),
                feature_library=ps.PolynomialLibrary(degree=3),
            )

        else:
            self._model = self._usr_sindy
        self._model.fit(
            [xi.states for xi in X],
            t=[xi.times for xi in X],
            multiple_trajectories=True,
        )

    def predict(self, iv: np.ndarray, times: np.ndarray) -> atraj.Trajectory:
        ret = self._model.simulate(iv, times)
        return atraj.Trajectory(times, ret, self._model.feature_names)

    @property
    def model(self) -> asys.ContinuousSystem:
        def gradient_f(t, x):
            return self._model.predict(np.atleast_2d(x))

        return asys.GradientContinuousSystem(gradient_f, self._model.feature_names)



# TODO: update this
__author__ = "Ethan Lew"
__copyright__ = "Copyright 2023"
__credits__ = [
    "Ethan Lew",
    "Abdelrahman Hekal",
    "Kostiantyn Potomkin",
    "Niklas Kochdumper",
    "Brandon Hencey",
    "Stanley Bak",
    "Sergiy Bogomolov",
]
__license__ = "GPLv3"
__maintainer__ = "Ethan Lew"
__email__ = "ethanlew16@gmail.com"
__status__ = "Prototype"

# we auto-manage versions
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from autokoopman.autokoopman import auto_koopman

from autokoopman.core.system import (
    ContinuousSystem,
    DiscreteSystem,
    GradientContinuousSystem,
    SymbolicContinuousSystem,
)
from autokoopman.core.trajectory import (
    TrajectoriesData,
    Trajectory,
    UniformTimeTrajectoriesData,
    UniformTimeTrajectory,
)

from . import _version

__version__ = _version.get_versions()["version"]



"""
Model has:

* a set of hyperparameters
"""
import abc
from typing import Optional

from sklearn.preprocessing import MinMaxScaler

import numpy as np

import autokoopman.core.system as asys
import autokoopman.core.trajectory as atraj


class TrajectoryEstimator(abc.ABC):
    """
    Trajectory based estimator base class

    :param normalize: apply MinMax normalization to the fit data
    :param feature_range: range for MinMax scaler
    """
    def __init__(self, normalize=True, feature_range=(-1, 1)) -> None:
        self.normalize = normalize
        if self.normalize:
            self.scaler = MinMaxScaler(feature_range=feature_range)
        else:
            self.scaler = None

    @abc.abstractmethod
    def fit(self, X: atraj.TrajectoriesData) -> None:
        pass

    @property
    @abc.abstractmethod
    def model(self) -> asys.System:
        pass


class NextStepEstimator(TrajectoryEstimator):
    """Estimator of discrete time dynamical systems

    Requires that the data be uniform time

    :param normalize: apply MinMax normalization to the fit data
    :param feature_range: range for MinMax scaler
    """

    def __init__(self, normalize=True, feature_range=(-1, 1)) -> None:
        super().__init__(normalize=normalize, feature_range=feature_range)
        self.names = None

    @abc.abstractmethod
    def fit_next_step(
        self, X: np.ndarray, Y: np.ndarray, U: Optional[np.ndarray] = None
    ) -> None:
        """an alternative fit method that uses a trajectories data structure"""
        pass

    def fit(self, X: atraj.TrajectoriesData) -> None:
        assert isinstance(
            X, atraj.UniformTimeTrajectoriesData
        ), "X must be uniform time"
        X_, Xp_, U_ = X.next_step_matrices
        if self.normalize:
            X_ = self.scaler.fit_transform(X_.T).T
            Xp_ = self.scaler.transform(Xp_.T).T
        self.fit_next_step(X_, Xp_, U_)
        self.sampling_period = X.sampling_period
        self.names = X.state_names


class GradientEstimator(TrajectoryEstimator):
    """Estimator of discrete time dynamical systems

    Requires that the data be uniform time
    
    :param normalize: apply MinMax normalization to the fit data
    :param feature_range: range for MinMax scaler
    """

    def __init__(self, normalize=True, feature_range=(-1, 1)) -> None:
        super().__init__(normalize=normalize, feature_range=feature_range)
        self.names = None

    @abc.abstractmethod
    def fit_gradient(
        self, X: np.ndarray, Y: np.ndarray, U: Optional[np.ndarray] = None
    ) -> None:
        """an alternative fit method that uses a trajectories data structure"""
        pass

    def fit(self, X: atraj.TrajectoriesData) -> None:
        assert isinstance(
            X, atraj.UniformTimeTrajectoriesData
        ), "X must be uniform time"
        X_, Xp_, U_ = X.differentiate
        if self.normalize:
            X_ = self.scaler.fit_transform(X_.T).T
            Xp_ = self.scaler.transform(Xp_.T).T
        self.fit_gradient(X_, Xp_, U_)
        self.sampling_period = X.sampling_period
        self.names = X.state_names



import sympy as sp  # type: ignore

import autokoopman.core.system as asys


class FitzHughNagumo(asys.SymbolicContinuousSystem):
    r"""
    FitzHugh-Nagumo System (FHN)
        The FHN Model is an example of a relaxation oscillator.

        .. math::
            \left\{\begin{array}{l}
            \dot{x}_{1}=x_0 - \frac{x_0^3}{3} - x_1 + R I_{ext} \\
            \dot{x}_{2}=\frac{x_0 + a - b x_1}{\tau} \\
            \end{array}\right.

    :param i_ext: external stimulus
    :param r: resistance
    :param a: model parameter :math:`a`
    :param b: model parameter :math:`b`
    :param tau: model parameter :math:`\tau`

    Setting :math:`a=b=0` creates the Van der Pol Oscillator.

    Example:
        .. code-block:: python

            import autokoopman.benchmark.fhn as fhn

            sys = fhn.FitzHughNagumo()
            traj = sys.solve_ivp(
                initial_state=[2.0, 1.5],
                tspan=(0.0, 20.0),
                sampling_period=0.01
            )
            traj.states
            # array([[ 2.        ,  1.5       ],
            #        [ 1.97862133,  1.50612782],
            #        [ 1.95779684,  1.51217922],
            #        ...,
            #        [-1.32849614, -0.81944446],
            #        [-1.32576847, -0.82204749],
            #        [-1.32303553, -0.82463882]])

    References
        FitzHugh, R. (1961). Impulses and physiological states in theoretical models of nerve membrane.
        Biophysical journal, 1(6), 445-466.
    """

    def __init__(self, i_ext=0.0, r=1.0, a=0.3, b=0.3, tau=3.0):
        x0, x1 = sp.symbols("x0 x1")
        xdot = [x0 - x0**3 / 3.0 - x1 + r * i_ext, (x0 + a - b * x1) / tau]
        super(FitzHughNagumo, self).__init__((x0, x1), xdot)



import autokoopman.benchmark.fhn as afhn
import numpy as np


def test_continuous():
    fhn = afhn.FitzHughNagumo()
    training_ivs = np.random.random((10, 2))

    # case test uniform time span
    ret_single = fhn.solve_ivp(training_ivs[0], tspan=(0.0, 3.0), sampling_period=0.05)
    ret_batch = fhn.solve_ivps(training_ivs, tspan=(0.0, 3.0), sampling_period=0.05)
    assert ret_batch.n_trajs == len(training_ivs)

    # case test teval
    teval = np.linspace(0.0, 10.0, 20)
    ret_batch = fhn.solve_ivps(training_ivs, teval=teval, sampling_period=0.05)
    ret_single = fhn.solve_ivp(training_ivs[0], teval=teval, sampling_period=0.05)
    assert len(ret_single.times) == len(teval)
    assert ret_batch.n_trajs == len(training_ivs)


def test_continuous_inputs():
    from autokoopman.benchmark.pendulum import PendulumWithInput

    iv = np.array([0.1, 0.0])
    teval = np.linspace(0.0, 5.0, 50)
    inputs = np.zeros(teval.shape)
    inputs[25] = 1.0
    pend = PendulumWithInput()
    sol = pend.solve_ivp(iv, teval=teval, inputs=inputs)


def test_discrete():
    from autokoopman.core.system import StepDiscreteSystem

    step = lambda t, s, i: np.array([s[0] - 0.1 * s[1], s[1] + 0.01 * s[0]])
    my_sys = StepDiscreteSystem(step, ["x0", "x1"])
    my_sys.solve_ivp(np.array([0.1, 0.1]), tspan=(0.0, 1.0), sampling_period=0.05)


def test_discrete_inputs():
    from autokoopman.core.system import StepDiscreteSystem

    step = lambda t, s, i: np.array([s[0] - 0.1 * s[1], s[1] + 0.01 * s[0] + i[0]])
    my_sys = StepDiscreteSystem(step, ["x0", "x1"])
    teval = np.linspace(0.0, 2.0, 20)
    inputs = teval
    my_sys.solve_ivp(
        np.array([0.1, 0.1]), teval=teval, inputs=inputs, sampling_period=0.05
    )



import abc
from typing import Callable, Sequence, Tuple, Optional, Union

import numpy as np
import scipy.integrate as scint  # type: ignore
import sympy as sp  # type: ignore

import autokoopman.core.trajectory as atraj
from autokoopman.core.format import _clip_list


class System(abc.ABC):
    @abc.abstractmethod
    def solve_ivp(
        self,
        initial_state: np.ndarray,
        tspan: Optional[Tuple[float, float]] = None,
        teval: Optional[np.ndarray] = None,
        inputs: Optional[np.ndarray] = None,
        sampling_period: float = 0.1,
    ) -> Union[atraj.Trajectory, atraj.UniformTimeTrajectory]:
        """
        Solve the initial value (IV) problem
            Given a system with a state space, specify how the system evolves with time given the initial conditions
            of the problem. The solution of a particular IV returns a trajectory over time teval if specified, or
            uniformly sampled over a time span given a sampling period.

        :param initial_state: IV state of system
        :param tspan: (if no teval is set) timespan to evolve system from iv (starting at tspan[0])
        :param teval: (optional) values of time sample the IVP trajectory
        :param sampling_period: (if no teval is set) sampling period of solution

        :returns: TimeTrajectory if teval is set, or UniformTimeTrajectory if not
        """
        raise NotImplementedError

    def solve_ivps(
        self,
        initial_states: np.ndarray,
        tspan: Optional[Tuple[float, float]] = None,
        teval: Optional[np.ndarray] = None,
        inputs: Optional[np.ndarray] = None,
        sampling_period: float = 0.1,
    ) -> Union[atraj.UniformTimeTrajectoriesData, atraj.TrajectoriesData]:
        ret = {}
        for idx, state in enumerate(initial_states):
            ret[idx] = self.solve_ivp(
                state,
                tspan=tspan,
                teval=teval,
                inputs=inputs[idx] if inputs is not None else None,
                sampling_period=sampling_period,
            )
        return atraj.UniformTimeTrajectoriesData(ret) if teval is None else atraj.TrajectoriesData(ret)  # type: ignore

    @property
    @abc.abstractmethod
    def names(self) -> Sequence[str]:
        pass

    @property
    def dimension(self) -> int:
        return len(self.names)

    def __repr__(self):
        return f"<{self.__class__.__name__} Dimensions: {self.dimension} States: {_clip_list(self.names)}>"


class ContinuousSystem(System):
    """
    Continuous Time System
        In this case, a CT system is a system whose evolution function is defined by a gradient.
    """

    def solve_ivp(
        self,
        initial_state: np.ndarray,
        tspan: Optional[Tuple[float, float]] = None,
        teval: Optional[np.ndarray] = None,
        inputs: Optional[np.ndarray] = None,
        sampling_period: float = 0.1,
    ) -> Union[atraj.Trajectory, atraj.UniformTimeTrajectory]:
        """
        Solve the initial value (IV) problem for Continuous Time Systems
            Given a system with a state space, specify how the system evolves with time given the initial conditions
            of the problem. The solution of a particular IV returns a trajectory over time teval if specified, or
            uniformly sampled over a time span given a sampling period.

            A differential equation of the form :math:`\dot X_t = \operatorname{grad}(t, X_t)`.

        :param initial_state: IV state of system
        :param tspan: (if no teval is set) timespan to evolve system from iv (starting at tspan[0])
        :param teval: (optional) values of time sample the IVP trajectory
        :param sampling_period: (if no teval is set) sampling period of solution

        :returns: TimeTrajectory if teval is set, or UniformTimeTrajectory if not
        """
        if teval is None and tspan is None:
            raise RuntimeError(f"teval or tspan must be set")
        if inputs is None:
            if teval is None:
                t_eval = np.arange(
                    tspan[0], tspan[-1] + sampling_period * (1 - 1e-12), sampling_period
                )
                sol = scint.solve_ivp(
                    self.gradient,
                    (min(t_eval), max(t_eval)),
                    initial_state,
                    args=(None,),
                    t_eval=t_eval,
                )
                return atraj.UniformTimeTrajectory(
                    sol.y.T,
                    None,
                    sampling_period,
                    state_names=self.names,
                    start_time=tspan[0],
                )
            else:
                sol = scint.solve_ivp(
                    self.gradient,
                    (min(teval), max(teval)),
                    initial_state,
                    args=(None,),
                    # TODO: this is hacky
                    t_eval=teval,
                )
                return atraj.Trajectory(sol.t, sol.y.T, None, self.names)
        else:
            if teval is not None:
                if len(teval) == 0:
                    raise ValueError("teval must have at least one value")
                inputs = np.array(inputs)
                sol = [initial_state]
                if len(teval) > 1:
                    for tcurrent, tnext, inpi in zip(
                        teval[:-1], teval[1:], inputs[:-1]
                    ):
                        sol_next = scint.solve_ivp(
                            self.gradient,
                            (tcurrent, tnext),
                            sol[-1],
                            args=(np.atleast_1d(inpi),),
                            t_eval=(tcurrent, tnext),
                        )
                        sol.append(sol_next.y.T[-1])
                if len(inputs.shape) == 1:
                    inputs = inputs[:, np.newaxis]
                return atraj.Trajectory(
                    np.array(teval), np.array(sol), inputs, self.names
                )
            else:
                raise RuntimeError("teval must be set if inputs is set")

    def solve_ivps(
        self,
        initial_states: np.ndarray,
        tspan: Optional[Tuple[float, float]] = None,
        teval: Optional[np.ndarray] = None,
        inputs: Optional[np.ndarray] = None,
        sampling_period: float = 0.1,
    ) -> Union[atraj.UniformTimeTrajectoriesData, atraj.TrajectoriesData]:
        ret = {}
        if inputs is not None:
            assert len(inputs) == len(
                initial_states
            ), f"length of inputs {len(inputs)} must match length of initial states {len(initial_states)}"
        for idx, state in enumerate(initial_states):
            ret[idx] = self.solve_ivp(
                state,
                tspan=tspan,
                teval=teval,
                inputs=inputs[idx] if inputs is not None else None,
                sampling_period=sampling_period,
            )
        return atraj.UniformTimeTrajectoriesData(ret) if teval is None else atraj.TrajectoriesData(ret)  # type: ignore

    @abc.abstractmethod
    def gradient(
        self, time: float, state: np.ndarray, sinput: Optional[np.ndarray]
    ) -> np.ndarray:
        raise NotImplementedError


class DiscreteSystem(System):
    """
    Discrete Time System
        In this case, a CT system is a system whose evolution function is defined by a next step function. For IVP, the
        discrete time can be related to continuous time via a sampling period. This trajectory can be interpolated to
        evaluate time points nonuniformly.

        TODO: should this have a sampling period instance member?
    """

    def solve_ivp(
        self,
        initial_state: np.ndarray,
        tspan: Optional[Tuple[float, float]] = None,
        teval: Optional[np.ndarray] = None,
        inputs: Optional[np.ndarray] = None,
        sampling_period: float = 0.1,
    ) -> Union[atraj.Trajectory, atraj.UniformTimeTrajectory]:
        """
        Solve the initial value (IV) problem for Discrete Time Systems
            Given a system with a state space, specify how the system evolves with time given the initial conditions
            of the problem. The solution of a particular IV returns a trajectory over time teval if specified, or
            uniformly sampled over a time span given a sampling period.

            A difference equation of the form :math:`X_{t+1} = \operatorname{step}(t, X_t)`.

        :param initial_state: IV state of system
        :param tspan: (if no teval is set) timespan to evolve system from iv (starting at tspan[0])
        :param teval: (optional) values of time sample the IVP trajectory
        :param sampling_period: (if no teval is set) sampling period of solution

        :returns: TimeTrajectory if teval is set, or UniformTimeTrajectory if not
        """
        if teval is None and tspan is None:
            raise RuntimeError(f"teval or tspan must be set")
        if inputs is None:
            if teval is None:
                times = np.arange(tspan[0], tspan[1] + sampling_period, sampling_period)
                states = np.zeros((len(times), len(self.names)))
                states[0] = np.array(initial_state).flatten()
                for idx, time in enumerate(times[1:]):
                    states[idx + 1] = self.step(
                        float(time), states[idx], None
                    ).flatten()
                return atraj.UniformTimeTrajectory(
                    states,
                    None,
                    sampling_period,
                    state_names=self.names,
                    start_time=tspan[0],
                )
            else:
                times = np.arange(
                    min(teval), max(teval) + sampling_period, sampling_period
                )
                states = np.zeros((len(times), len(self.names)))
                states[0] = np.array(initial_state).flatten()
                for idx, time in enumerate(times[1:]):
                    states[idx + 1] = self.step(
                        float(time), states[idx], None
                    ).flatten()
                traj = atraj.Trajectory(times, states, None, self.names)
                return traj.interp1d(teval)
        else:
            if teval is not None:
                if len(teval) == 0:
                    raise ValueError("teval must have at least one value")
                if inputs.ndim == 1:
                    inputs = inputs[:, np.newaxis]
                teval = np.array(teval)
                times = np.arange(
                    min(teval), max(teval) + sampling_period, sampling_period
                )
                states = np.zeros((len(times), len(self.names)))
                states[0] = np.array(initial_state).flatten()
                for idx, time in enumerate(times[1:]):
                    diff = time - teval
                    diff[diff < 0.0] = float("inf")
                    tidx = diff.argmin()
                    states[idx + 1] = self.step(
                        float(time), states[idx], np.atleast_1d(inputs[tidx])
                    ).flatten()
                traj = atraj.Trajectory(times, states, None, state_names=self.names)
                ctraj = traj.interp1d(teval)
                return atraj.Trajectory(teval, ctraj.states, inputs, self.names)
            else:
                raise RuntimeError("teval must be set if inputs is set")

    @abc.abstractmethod
    def step(
        self, time: float, state: np.ndarray, sinput: Optional[np.ndarray]
    ) -> np.ndarray:
        raise NotImplementedError


class SymbolicContinuousSystem(ContinuousSystem):
    def __init__(
        self,
        variables: Sequence[sp.Symbol],
        gradient_exprs: Sequence[sp.Expr],
        input_variables: Optional[Sequence[sp.Symbol]] = None,
        time_var=None,
    ):
        if time_var is None:
            time_var = sp.symbols("_t0")
        if input_variables is None:
            self._variables = [time_var, *variables]
        else:
            self._variables = [time_var, *variables, *input_variables]
        self._state_vars = variables
        self._input_vars = input_variables
        self._exprs = gradient_exprs
        self._mat = sp.Matrix(self._exprs)
        self._fmat = sp.lambdify((self._variables,), self._mat)

    def gradient(
        self, time: float, state: np.ndarray, sinput: Optional[np.ndarray]
    ) -> np.ndarray:
        if sinput is None:
            return np.array(self._fmat(np.array([time, *state]))).flatten()
        else:
            return np.array(self._fmat(np.array([time, *state, *sinput]))).flatten()

    @property
    def names(self) -> Sequence[str]:
        return [str(s) for s in self._state_vars]


class GradientContinuousSystem(ContinuousSystem):
    def __init__(
        self,
        gradient_func: Callable[[float, np.ndarray, Optional[np.ndarray]], np.ndarray],
        names,
    ):
        self._names = names
        self._gradient_func = gradient_func

    def gradient(
        self, time: float, state: np.ndarray, sinput: Optional[np.ndarray]
    ) -> np.ndarray:
        return self._gradient_func(time, state, sinput)

    @property
    def names(self):
        return self._names


class StepDiscreteSystem(DiscreteSystem):
    def __init__(
        self,
        step_func: Callable[[float, np.ndarray, Optional[np.ndarray]], np.ndarray],
        names,
    ):
        self._names = names
        self._step_func = step_func

    def step(
        self, time: float, state: np.ndarray, sinput: Optional[np.ndarray]
    ) -> np.ndarray:
        return self._step_func(time, state, sinput)

    @property
    def names(self):
        return self._names


class LinearContinuousSystem(ContinuousSystem):
    ...


class KoopmanContinuousSystem(LinearContinuousSystem):
    ...


class KoopmanSystem:
    """
    mixin class for Koopman systems
    """

    def __init__(self, A, B, obs, names, dim=None, scaler=None):
        self._A = A
        self._B = B
        self._has_input = B is not None and not np.any(np.array(B.shape) == 0)
        self.obs = obs
        self.scaler = scaler
        self.dim = A.shape[0] if dim is None else dim

        def evolv_func(t, x, i):
            obs = (self.obs(x).flatten())[np.newaxis, :]
            if self._has_input:
                return np.real(
                    self._A @ obs.T + self._B @ (i)[:, np.newaxis]
                ).flatten()[: self.dim]
            else:
                return np.real(self._A @ obs.T).flatten()[: len(x)]

        def evolv_func_scale(t, x, i):
            x = self.scaler.transform(np.atleast_2d(x)).flatten()
            obs = (self.obs(x).flatten())[np.newaxis, :]
            if self._has_input:
                return self.scaler.inverse_transform(
                    np.real(self._A @ obs.T + self._B @ (i)[:, np.newaxis])[
                        : self.dim
                    ].T
                ).flatten()
            else:
                return self.scaler.inverse_transform(
                    np.real(self._A @ obs.T)[: len(x)].T
                ).flatten()

        super().__init__(evolv_func if self.scaler is None else evolv_func_scale, names)

    @property
    def A(self):
        return self._A

    @property
    def B(self):
        return self._B

    @property
    def obs_func(self):
        return self.obs


class KoopmanStepDiscreteSystem(KoopmanSystem, StepDiscreteSystem):
    ...


class KoopmanGradientContinuousSystem(KoopmanSystem, GradientContinuousSystem):
    ...



import numpy as np
import pytest

koop_config = ((True,), (False,))


@pytest.mark.parametrize(
    "normalize",
    koop_config,
)
def test_discrete_koopman(normalize):
    """tests the discrete time Koopman estimator"""
    from autokoopman.benchmark.pendulum import PendulumWithInput
    from autokoopman.core.trajectory import TrajectoriesData
    from autokoopman.core.observables import IdentityObservable, RFFObservable
    from autokoopman.estimator.koopman import KoopmanDiscEstimator
    import random

    pend = PendulumWithInput(l=0.6, beta=0.1)
    n_trajs = 40
    training_ivs = np.random.random((n_trajs, 2))
    trajs = {}
    teval = np.linspace(0.0, 5.0, 100)
    for i in range(n_trajs):
        inputs = (
            random.random() * 2.0 * np.cos(random.random() * teval + random.random())
        )
        sol = pend.solve_ivp(training_ivs[i], teval=teval, inputs=inputs)
        trajs[i] = sol
    sols = TrajectoriesData(trajs)
    sols = sols.interp_uniform_time(1 / 20.0)
    disc = KoopmanDiscEstimator(
        IdentityObservable() | RFFObservable(2, 100, 0.001),
        0.1,
        2,
        100,
        normalize=normalize,
    )
    disc.fit(sols)
    new_inputs = inputs.copy()
    new_inputs[-len(inputs) // 2 :] = -10.0
    new_inputs[: len(inputs) // 2] = 0.0
    preds = disc.model.solve_ivps(
        training_ivs,
        teval=teval,
        inputs=[new_inputs for _ in range(len(training_ivs))],
        sampling_period=1 / 20.0,
    )
    new_sols = pend.solve_ivps(
        training_ivs,
        teval=teval,
        inputs=[new_inputs for _ in range(len(training_ivs))],
        sampling_period=1 / 20.0,
    )


@pytest.mark.parametrize(
    "normalize",
    koop_config,
)
def test_cont_koopman(normalize):
    """tests the continuous time Koopman estimator"""
    from autokoopman.benchmark.pendulum import PendulumWithInput
    from autokoopman.core.trajectory import TrajectoriesData
    from autokoopman.core.observables import IdentityObservable, RFFObservable
    from autokoopman.estimator.koopman import KoopmanContinuousEstimator
    import random

    pend = PendulumWithInput(l=0.6, beta=0.1)
    n_trajs = 40
    training_ivs = np.random.random((n_trajs, 2))
    trajs = {}
    teval = np.linspace(0.0, 5.0, 100)
    for i in range(n_trajs):
        inputs = (
            random.random() * 2.0 * np.cos(random.random() * teval + random.random())
        )
        sol = pend.solve_ivp(training_ivs[i], teval=teval, inputs=inputs)
        trajs[i] = sol
    sols = TrajectoriesData(trajs)
    sols = sols.interp_uniform_time(1 / 20.0)
    cont = KoopmanContinuousEstimator(
        IdentityObservable() | RFFObservable(2, 100, 0.001), 2, 100, normalize=normalize
    )
    cont.fit(sols)
    new_inputs = inputs.copy()
    new_inputs[-len(inputs) // 2 :] = -10.0
    new_inputs[: len(inputs) // 2] = 0.0
    preds = cont.model.solve_ivps(
        training_ivs,
        teval=teval,
        inputs=[new_inputs for _ in range(len(training_ivs))],
        sampling_period=1 / 20.0,
    )
    new_sols = pend.solve_ivps(
        training_ivs,
        teval=teval,
        inputs=[new_inputs for _ in range(len(training_ivs))],
        sampling_period=1 / 20.0,
    )



import numpy as np
from pydmd import DMD  # type: ignore
from pydmd.dmd import compute_tlsq  # type: ignore

import autokoopman.core.estimator as aest
import autokoopman.core.system as asys


class MultiDMD(DMD):
    def fit_multi(self, X, Y):
        """
        Compute the Dynamic Modes Decomposition to the input data.
        :param X: the input snapshots.
        :type X: numpy.ndarray or iterable
        """
        self._snapshots, self._snapshots_shape = self._col_major_2darray(X)

        # n_samples = self._snapshots.shape[1]
        # X = self._snapshots[:, :-1]
        # Y = self._snapshots[:, 1:]

        X, Y = compute_tlsq(X, Y, self.tlsq_rank)
        self._svd_modes, _, _ = self.operator.compute_operator(X, Y)

        # Default timesteps
        # self._set_initial_time_dictionary(
        #    {"t0": 0, "tend": n_samples - 1, "dt": 1}
        # )

        self._b = self._compute_amplitudes()

        return self


class DMDEstimator(aest.NextStepEstimator):
    def fit_next_step(self, X: np.ndarray, Y: np.ndarray) -> None:
        self.dmd = MultiDMD(svd_rank=-1)
        self.dmd.fit_multi(X, Y)

    @property
    def model(self) -> asys.System:
        def step_func(t, x):
            return np.real(self.dmd.predict(np.atleast_2d(x).T)).T

        return asys.StepDiscreteSystem(step_func, self.names)



import sympy as sp  # type: ignore

import autokoopman.core.system as asys


class RobBench(asys.SymbolicContinuousSystem):
    r"""
    Robertson chemical reaction benchmark (ROBE21)
        As proposed by Robertson [31], this chemical reaction system models the kinetics of an auto-catalytic reaction.
        x, y and z are the (positive) concentrations of the species, with the assumption that x+y+z = 1. Here alpha is a small
        constant, while beta and gamma take on large values.
        The initial condition is always :math:`x(0) = 1`, :math:`y(0) = 0` and :math:`z(0) = 0`.

        :math:`I: x_0(0) \in [9.5, 10.0]`, i.e., uncertainty on the initial condition;

        In this benchmark we fix alpha = 0.4 and analyze the system under three different pairs of values for beta and gamma:
        1. beta = 10^2 , gamma = 10^3
        2. beta=10^3, gamma = 10^5
        3. beta = 10^3, gamma = 10^7

        We are interested in computing the reachable tube until t = 40, to see how the integration scheme holds under the stiff behavior. No verification objective is enforced.

        For each of the three setups, the following three measures are required:
        1. the execution time for evolution;
        2. the number of integration steps taken;
        3. the width of the sum of the concentrations s = x + y + z at the final time.
        Additionally, a figure with s (in the [0.999, 1.001] range) w.r.t. time overlaid for the three setups should be shown.

        As proposed by Robertson, this chemical reaction system models the kinetics of an auto- catalytic reaction.

    References
        Geretti, L., dit Sandretto, J. A., Althoff, M., Benet, L., Chapoutot, A., Collins, P., ... & Wetzlinger, M. (2021). ARCH-COMP21 Category Report: Continuous and Hybrid Systems with Nonlinear Dynamics. EPiC Series in Computing, 80, 32-54.
        H. H. Robertson. The solution of a set of reaction rate equations. In ”Numerical analysis: an introduction”, page 178–182. Academic Press, 1966.
    """

    def __init__(self, alpha=0.4, beta=100, gamma=1000):
        x, y, z = sp.symbols("x y z")
        xdot = [
            -alpha * x + beta * y * z,
            alpha * x - beta * y * z - gamma * y * y,
            gamma * y * y,
        ]
        super(RobBench, self).__init__((x, y, z), xdot)



"""
Main AutoKoopman Function (Convenience Function)
"""
from typing import Callable, Union, Sequence, Optional, Tuple

import numpy as np

import autokoopman.core.observables as kobs
from autokoopman.core.trajectory import (
    TrajectoriesData,
    UniformTimeTrajectoriesData,
    UniformTimeTrajectory,
)
from autokoopman.core.tuner import (
    HyperparameterTuner,
    HyperparameterMap,
    ParameterSpace,
    ContinuousParameter,
    DiscreteParameter,
    TrajectoryScoring,
)
from autokoopman.estimator.koopman import KoopmanDiscEstimator
from autokoopman.tuner.gridsearch import GridSearchTuner
from autokoopman.tuner.montecarlo import MonteCarloTuner
from autokoopman.tuner.bayesianopt import BayesianOptTuner
from autokoopman.core.observables import KoopmanObservable
from autokoopman.core.format import hide_prints

__all__ = ["auto_koopman"]


# valid string identifiers for the autokoopman magic
obs_types = {"rff", "poly", "quadratic", "id", "deep"}
opt_types = {"grid", "monte-carlo", "bopt"}
scoring_func_types = {"total", "end", "relative"}


def get_scoring_func(score_name):
    """resolve scoring function from name or callable type"""
    # if callable, just return it
    if callable(score_name):
        return score_name
    if score_name == "total":
        return TrajectoryScoring.total_score
    elif score_name == "end":
        return TrajectoryScoring.end_point_score
    elif score_name == "relative":
        return TrajectoryScoring.relative_score
    else:
        raise ValueError(
            f"Scoring function name {score_name} is not in available list (names are {scoring_func_types})"
        )


def get_parameter_space(obs_type, threshold_range, rank):
    """from the myriad of user suppled switches, select the right hyperparameter space"""
    if obs_type == "rff":
        return ParameterSpace(
            "koopman-rff",
            [
                ContinuousParameter(
                    "gamma", *threshold_range, distribution="loguniform"
                ),
                DiscreteParameter("rank", *rank),
            ],
        )
    elif obs_type == "quadratic":
        return ParameterSpace(
            "koopman-quadratic",
            [
                DiscreteParameter("rank", *rank),
            ],
        )
    elif obs_type == "poly":
        return ParameterSpace(
            "koopman-polynomial",
            [
                DiscreteParameter("degree", 1, 5),
                DiscreteParameter("rank", *rank),
            ],
        )

    elif obs_type == "id":
        return ParameterSpace(
            "koopman-id",
            [
                DiscreteParameter("rank", *rank),
            ],
        )
    elif isinstance(obs_type, KoopmanObservable):
        # assume user has set everything else
        return ParameterSpace(
            "koopman-id",
            [
                DiscreteParameter("rank", *rank),
            ],
        )


def get_estimator(obs_type, sampling_period, dim, obs, hyperparams, normalize):
    """from the myriad of user suppled switches, select the right estimator"""
    if obs_type == "rff":
        observables = kobs.IdentityObservable() | kobs.RFFObservable(
            dim, obs, hyperparams[0]
        )
        return KoopmanDiscEstimator(
            observables, sampling_period, dim, rank=hyperparams[1], normalize=normalize
        )
    elif obs_type == "quadratic":
        observables = kobs.IdentityObservable() | kobs.QuadraticObservable(dim)
        return KoopmanDiscEstimator(
            observables, sampling_period, dim, rank=hyperparams[0], normalize=normalize
        )
    elif obs_type == "poly":
        observables = kobs.PolynomialObservable(dim, hyperparams[0])
        return KoopmanDiscEstimator(
            observables, sampling_period, dim, rank=hyperparams[1], normalize=normalize
        )
    elif obs_type == "id":
        observables = kobs.IdentityObservable()
        return KoopmanDiscEstimator(
            observables, sampling_period, dim, rank=hyperparams[0], normalize=normalize
        )
    elif isinstance(obs_type, KoopmanObservable):
        return KoopmanDiscEstimator(
            obs_type, sampling_period, dim, rank=hyperparams[0], normalize=normalize
        )
    else:
        raise ValueError(f"unknown observables type {obs_type}")


def auto_koopman(
    training_data: Union[TrajectoriesData, Sequence[np.ndarray]],
    inputs_training_data: Optional[Sequence[np.ndarray]] = None,
    sampling_period: Optional[float] = None,
    normalize: bool = True,
    opt: Union[str, HyperparameterTuner] = "monte-carlo",
    max_opt_iter: int = 100,
    max_epochs: int = 500,
    n_splits: Optional[int] = None,
    obs_type: Union[str, KoopmanObservable] = "rff",
    cost_func: Union[
        str, Callable[[TrajectoriesData, TrajectoriesData], float]
    ] = "total",
    n_obs: int = 100,
    rank: Optional[Union[Tuple[int, int], Tuple[int, int, int]]] = None,
    grid_param_slices: int = 10,
    lengthscale: Tuple[float, float] = (1e-4, 1e1),
    enc_dim: Tuple[int, int, int] = (2, 64, 16),
    n_layers: Tuple[int, int, int] = (1, 8, 2),
    torch_device: Optional[str] = None,
    verbose: bool = True,
):
    """
    AutoKoopman Convenience Function
        This is an interface to the dynamical systems learning functionality of the AutoKoopman library. The user can select
        estimators classes at a high level. A tuner can be chosen to find the best hyperparameter values.

    :param training_data: training trajectories data from which to learn the system
    :param inputs_training_data: optional input trajectories data from which to learn the system (this isn't needed if the training data has inputs already)
    :param sampling_period: (for discrete time system) sampling period of training data
    :param normalize: normalize the states of the training trajectories
    :param opt: hyperparameter optimizer {"grid", "monte-carlo", "bopt"}
    :param max_opt_iter: maximum iterations for the tuner to use
    :param max_epochs: maximum number of training epochs
    :param n_splits: (for optimizers) if set, switches to k-folds bootstrap validation for the hyperparameter tuning. This is useful for things like RFF tuning where the results have noise.
    :param obs_type: (for koopman) koopman observables to use {"rff", "quadratic", "poly", "id", "deep"}
    :param cost_func: cost function to use for hyperparameter optimization
    :param  n_obs: (for koopman) number of observables to use (if applicable)
    :param rank: (for koopman) rank range (start, stop) or (start, stop, step)
    :param grid_param_slices: (for grid tuner) resolution to slice continuous valued parameters into
    :param lengthscale: (for RFF observables) RFF kernel lengthscale
    :param enc_dim: (for deep learning) number of dimensions in the latent space
    :param n_layers: (for deep learning) number of hidden layers in the encoder / decoder
    :param torch_device: (for deep learning) specify torch compute device
    :param verbose: whether to print progress and messages

    :returns: Tuned Model and Metadata

    Example:
        .. code-block:: python

            from autokoopman.benchmark.fhn import FitzHughNagumo
            from autokoopman import auto_koopman

            # let's build an example dataset
            data = fhn.solve_ivps(
                initial_states=[[0.0, -4.0], [1.0, 3.4], [1.0, 1.0], [0.1, -0.1]],
                tspan=[0.0, 1.0], sampling_period=0.01
            )

            # learn a system
            results = auto_koopman(
                data,
                obs_type="rff",
                opt="grid",
                n_obs=200,
                max_opt_iter=200,
                grid_param_slices=10,
                n_splits=3,
                rank=(1, 200, 20)
            )

            # results = {'tuned_model': <StepDiscreteSystem Dimensions: 2 States: [X1, X2]>,
            # 'model_class': 'koopman-rff',
            # 'hyperparameters': ['gamma', 'rank'],
            # 'hyperparameter_values': (0.004641588833612782, 21),
            # 'tuner_score': 0.14723275426562,
            # 'tuner': <autokoopman.tuner.gridsearch.GridSearchTuner at 0x7f0f92f95580>,
            # 'estimator': <autokoopman.estimator.koopman.KoopmanDiscEstimator at 0x7f0f92ff0610>}
    """

    training_data, sampling_period = _sanitize_training_data(
        training_data, inputs_training_data, sampling_period, opt, obs_type
    )

    # get the hyperparameter map
    if obs_type in {"deep"}:
        modelmap = _deep_model_map(
            training_data,
            max_epochs,
            n_obs,
            enc_dim,
            n_layers,
            torch_device,
            verbose,
            normalize,
        )
    else:
        modelmap = _edmd_model_map(
            training_data,
            rank,
            obs_type,
            n_obs,
            lengthscale,
            sampling_period,
            normalize,
        )

    # setup the tuner
    if opt == "grid":
        gt = GridSearchTuner(
            modelmap,
            training_data,
            n_samps=grid_param_slices,
            n_splits=n_splits,
            verbose=verbose,
        )
    elif opt == "monte-carlo":
        gt = MonteCarloTuner(
            modelmap, training_data, n_splits=n_splits, verbose=verbose
        )
    elif opt == "bopt":
        gt = BayesianOptTuner(modelmap, training_data, verbose=verbose)
    else:
        raise ValueError(f"could not match a tuner to the string {opt}")

    with hide_prints():
        res = gt.tune(nattempts=max_opt_iter, scoring_func=get_scoring_func(cost_func))

    # pack results into out custom output
    result = {
        "tuned_model": res["model"].model,
        "model_class": modelmap.parameter_space.name,
        "hyperparameters": [param.name for param in modelmap.parameter_space],
        "hyperparameter_values": res["param"],
        "tuner_score": res["score"],
        "tuner": gt,
        "estimator": res["model"],
        "has_errors": len(gt.error_messages) > 0,
        "error_messages": gt.error_messages,
    }
    return result


def _deep_model_map(
    training_data: TrajectoriesData,
    epochs,
    obs_dim,
    enc_dim,
    nlayers,
    torch_device,
    verbose,
    normalize,
) -> HyperparameterMap:
    import autokoopman.estimator.deepkoopman as dk

    pspace = ParameterSpace(
        "koopman-deep",
        [
            # DiscreteParameter("obs_dim", *obs_dim),
            DiscreteParameter("enc_dim", *enc_dim),
            DiscreteParameter("num_layers", *nlayers),
        ],
    )

    # system dimension
    dim = len(training_data.state_names)
    input_dim = (
        len(training_data.input_names) if training_data.input_names is not None else 0
    )

    class _ModelMap(HyperparameterMap):
        def __init__(self):
            self.names = training_data.state_names
            super(_ModelMap, self).__init__(pspace)

        def get_model(self, hyperparams: Sequence):
            return dk.DeepKoopman(
                state_dim=dim,
                input_dim=input_dim,
                hidden_dim=hyperparams[0],
                max_iter=epochs,
                lr=1e-3,
                hidden_enc_dim=64,
                num_hidden_layers=hyperparams[1],
                pred_loss_weight=1.0,
                metric_loss_weight=0.1,
                torch_device=torch_device,
                # turning this off because of review feedback
                verbose=False,
                display_progress=False,  # don't nest progress bars
            )

    # get the hyperparameter map
    return _ModelMap()


def _edmd_model_map(
    training_data, rank, obs_type, n_obs, lengthscale, sampling_period, normalize
) -> HyperparameterMap:
    """model map for eDMD based methods

    :param training_data:
    :param rank: set of ranks to try (of DMD rank parameter)
    :param obs_type:
    :param n_obs: some obs type require a number of observables
    :sampling_period:

    :returns: hyperparameter map
    """
    # system dimension
    dim = len(training_data.state_names)

    # infer rank range and step
    if rank is not None:
        if len(rank) == 2:
            rank = (*rank, 1)
        else:
            rank = rank
    else:
        rank = (2, n_obs + dim, 10)

    # get the hyperparameter space
    pspace = get_parameter_space(obs_type, lengthscale, rank)

    class _ModelMap(HyperparameterMap):
        """SINDy with hyperparameters for polynomial library"""

        def __init__(self):
            self.names = training_data.state_names
            super(_ModelMap, self).__init__(pspace)

        def get_model(self, hyperparams: Sequence):
            return get_estimator(
                obs_type, sampling_period, dim, n_obs, hyperparams, normalize
            )

    # get the hyperparameter map
    return _ModelMap()


def _sanitize_training_data(
    training_data, inputs_training_data, sampling_period, opt, obs_type
):
    """auto_koopman input sanitization"""

    # if sampling period is None AND discrete system is wanted
    if sampling_period is None:
        sampling_period = np.infty
        for t in training_data:
            sampling_period = min(sampling_period, min(np.diff(t.times)))

    # sanitize the input
    # check the strings
    if isinstance(obs_type, str):
        assert (
            obs_type in obs_types
        ), f"observable name {obs_type} is unknown (valid ones are {obs_types})"
    if isinstance(opt, str):
        assert (
            opt in opt_types
        ), f"tuner name {opt} is unknown (valid ones are {opt_types})"

    # convert the data to autokoopman trajectories
    if isinstance(training_data, TrajectoriesData):
        if not isinstance(training_data, UniformTimeTrajectoriesData):
            print(
                f"resampling trajectories as they need to be uniform time (sampling period {sampling_period})"
            )
            training_data = training_data.interp_uniform_time(sampling_period)
        else:
            if not np.isclose(training_data.sampling_period, sampling_period):
                print(
                    f"resampling trajectories because the sampling periods differ (original {training_data.sampling_period}, new {sampling_period})"
                )
                training_data = training_data.interp_uniform_time(sampling_period)
    else:
        # figure out how to add inputs
        training_iter = (
            training_data.items() if isinstance(training_data, dict) else training_data
        )
        if inputs_training_data is not None:
            training_iter = [(n, x, inputs_training_data[n]) for n, x in training_iter]
        else:
            training_iter = [(n, x, None) for n, x in training_iter]
        if isinstance(training_data, dict):
            training_data = UniformTimeTrajectoriesData(
                {
                    k: UniformTimeTrajectory(v, u, sampling_period=sampling_period)
                    for k, v, u in training_iter
                }
            )
        else:
            training_data = UniformTimeTrajectoriesData(
                {
                    idx: UniformTimeTrajectory(di, u, sampling_period=sampling_period)
                    for idx, di, u in training_iter
                }
            )

    return training_data, sampling_period



from typing import Dict, Hashable, Sequence, Set, Tuple, Union, List, Optional

import numpy as np
import pandas as pd

from autokoopman.core.format import _clip_list


class Trajectory:
    """a trajectory is a time series of vectors"""

    def __init__(
        self,
        times: np.ndarray,
        states: np.ndarray,
        inputs: Optional[np.ndarray],
        state_names: Optional[Sequence[str]] = None,
        input_names: Optional[Sequence[str]] = None,
        threshold=None,
    ):
        def _make_names(
            names: Optional[Sequence[str]],
            ref_in: Optional[np.ndarray],
            ref_char: str,
            category: str,
        ):
            if names is None:
                if ref_in is None:
                    return None
                else:
                    assert (
                        len(np.array(ref_in).shape) == 2
                    ), f"{category} must be able to be converted to 2D numpy array"
                    l = np.array(ref_in).shape[1]
                    return [f"{ref_char}{idx}" for idx in range(l)]
            else:
                return names

        assert times.ndim == 1, "times must be a 1d array"
        self._times = times
        self._states = states
        self._inputs = inputs
        self._state_names = _make_names(state_names, self._states, "x", "states")
        self._input_names = _make_names(input_names, self._inputs, "u", "inputs")
        self._threshold = 1e-12 if threshold is None else threshold

        # more checks...
        assert self.dimension == len(
            self.names
        ), "dimension must match the length of state names!"
        assert (
            self.input_dimension == len(self.input_names)
            if self.input_names is not None
            else True
        ), "input dimension must match the length of input names!"
        assert len(self.times) == len(
            self.states
        ), "length of times must match length of states"
        if input_names is not None and self.inputs is not None:
            assert len(self.times) == len(
                self.inputs
            ), "length of inputs must match length of times"

    @property
    def times(self) -> np.ndarray:
        return self._times

    @property
    def size(self) -> int:
        return self.states.shape[0]

    @property
    def dimension(self) -> int:
        return self.states.shape[1]

    @property
    def input_dimension(self) -> Optional[int]:
        if self.inputs is None:
            return None
        else:
            return self.inputs.shape[1]

    @property
    def states(self) -> np.ndarray:
        return self._states

    @property
    def inputs(self) -> Optional[np.ndarray]:
        return self._inputs

    @property
    def names(self) -> Sequence:
        return self._state_names

    @property
    def input_names(self) -> Optional[Sequence]:
        return self._input_names

    @property
    def is_uniform_time(self) -> bool:
        """determine if a trajectory is uniform time"""
        # FIXME: len greater than 3??
        time_steps = np.diff(self._times)
        dtimes = np.diff(time_steps)
        return bool(np.all([ti < self._threshold for ti in np.abs(dtimes)]))

    def interp1d(self, times) -> "Trajectory":
        from scipy.interpolate import interp1d  # type: ignore

        if self._inputs is None:
            times = np.array(times)
            f = interp1d(
                np.array(self._times),
                np.array(self._states),
                fill_value="extrapolate",
                axis=0,
            )
            states = f(times)
            return Trajectory(
                times, states, None, self.names, threshold=self._threshold
            )
        else:
            times = np.array(times)
            f = interp1d(
                np.array(self._times),
                np.array(self._states),
                fill_value="extrapolate",
                axis=0,
            )
            fi = interp1d(
                np.array(self._times),
                np.array(self._inputs),
                fill_value="extrapolate",
                axis=0,
            )
            states = f(times)
            inputs = fi(times)
            return Trajectory(
                times,
                states,
                inputs,
                state_names=self.names,
                input_names=self.input_names,
                threshold=self._threshold,
            )

    def interp_uniform_time(self, sampling_period) -> "UniformTimeTrajectory":
        ts = np.arange(
            np.min(self._times), np.max(self._times) + sampling_period, sampling_period
        )
        return self.interp1d(ts).to_uniform_time_traj()

    def to_uniform_time_traj(self) -> "UniformTimeTrajectory":
        assert self.is_uniform_time, "trajectory must be uniform time to apply"
        sp = np.diff(self._times)[0]
        start = self._times[0]
        return UniformTimeTrajectory(
            self.states,
            self.inputs,
            sp,
            state_names=self.names,
            input_names=self.input_names,
            start_time=start,
            threshold=self._threshold,
        )

    def __repr__(self):
        return f"<{self.__class__.__name__} Dim: {self.dimension}, Length: {self.size}, State Names: {_clip_list(self.names)}, Input Names: {_clip_list(self.input_names)}>"

    def _prepare_other(self, other: "Trajectory"):
        if len(self.times) != len(other.times) or not np.all(
            np.isclose(self.times, other.times)
        ):
            return other.interp1d(self.times)
        else:
            return other

    def __sub__(self, other: "Trajectory") -> "Trajectory":
        otheri = self._prepare_other(other)
        return Trajectory(
            self.times,
            self.states - otheri.states,
            None,
            state_names=self.names,
            input_names=None,
            threshold=self._threshold,
        )

    def __add__(self, other: "Trajectory") -> "Trajectory":
        otheri = self._prepare_other(other)
        return Trajectory(
            self.times,
            self.states + otheri.states,
            None,
            state_names=self.names,
            input_names=None,
            threshold=self._threshold,
        )

    def norm(self, axis=1) -> "Trajectory":
        # TODO: allow the norm type to be selected?
        from numpy.linalg import norm

        return Trajectory(
            self.times,
            norm(self.states, axis=axis)[:, np.newaxis],
            None,
            state_names=["<norm>"],
            input_names=None,
            threshold=self._threshold,
        )


class UniformTimeTrajectory(Trajectory):
    """uniform time is a trajectory advanced by a sampling period"""

    def __init__(
        self,
        states: np.ndarray,
        inputs: Optional[np.ndarray],
        sampling_period: float,
        state_names: Optional[Sequence[str]] = None,
        input_names: Optional[Sequence[str]] = None,
        start_time=0.0,
        threshold=None,
    ):
        nsteps = len(states)
        times = np.arange(
            start_time, start_time + nsteps * sampling_period, sampling_period
        )[:nsteps]
        super().__init__(
            times, states, inputs, state_names, input_names, threshold=threshold
        )
        self._sampling_period = sampling_period

    @property
    def sampling_period(self) -> float:
        """
        Sampling Period
            A sampling period allows the trajectories to relate to discrete and continuous time signals. Given a
            discrete time signal :math:`x_d`, it relates to an analog signal :math:`x_a` by a sampling time
            :math:`T_s`,

            .. math::

                x_a(n T_s) = x_d(n)
        """
        return self._sampling_period

    @property
    def sampling_frequency(self) -> float:
        return 1.0 / self.sampling_period

    @property
    def is_uniform_time(self) -> bool:
        """determine if a trajectory is uniform time"""
        return True


class TrajectoriesData:
    """
    A Dataset of Trajectories (Multiple Time Series Vectors)
        This is a data access object (DAO) for datasets used to learn dynamical systems from observed trajectories. The
        object supports

        * **state / trajectory identifiers** trajectories can be accessed and manipulated by name
        * **iteration** the trajectories can be iterated through
        * **serialization** trajectory datasets can be read and saved from CSV files and Pandas DataFrames
        * **interpolation** trajectories can be resampled in time (e.g. convert to uniform time for certain technique)
    """

    traj_id_hname = "id"
    time_id_hname = "time"

    @staticmethod
    def equal_lists(lists: Sequence[Sequence]):
        """
        List Equality
            In a sequence of lists, determine if all list are equal to one another.
        """
        return not lists or all(lists[0] == b for b in lists[1:])

    @classmethod
    def from_pandas(
        cls, data_df: pd.DataFrame, threshold=None, traj_id="id", time_id="time"
    ) -> "TrajectoriesData":
        """
        Create TrajectoriesData DAO from Pandas DataFrame
            To be a valid DataFrame the object must have a "time" field to record the sample time and an "id" field to distinguish
            the different trajectories.

        :param data_df: Pandas Dataframe to convert to trajectories.
        :param threshold: trajectories threshold value.
        :param traj_id: column name for trajectory identifier.
        :param time_id: column name for time identifier.

        :returns: TrajectoriesData

        Examples
            .. code-block:: python

                import autokoopman.core.trajectory as atraj
                import pandas as pd

                # create example dataframe
                t = np.linspace(0, 1.0, 10)
                my_df = pd.DataFrame(
                    columns=["time", "state", "id"],
                    data=np.array([
                        t,
                        np.sin(t),
                        np.ones(shape=t.shape)
                    ]).T
                )

                # create TrajectoriesData from the DataFrame
                trajs = atraj.TrajectoriesData.from_pandas(my_df)
        """
        traj_ids = set(data_df[traj_id])
        assert traj_id in set(data_df.columns) and time_id in set(data_df.columns), (
            f"csv must have {traj_id} " f"and {time_id} fields"
        )
        assert len(data_df.columns) > 2, "csv has to have more than two columns"
        tidx, sidx = (
            list(data_df.columns).index(time_id),
            list(data_df.columns).index(traj_id),
        )
        state_names = list(data_df.columns)[tidx + 1 : sidx]
        trajs = {
            uvi: Trajectory(
                data_df[data_df[traj_id] == uvi][time_id].to_numpy(),
                data_df[data_df[traj_id] == uvi][state_names].to_numpy(),
                None,
                state_names,
                None,
                threshold=threshold,
            )
            for uvi in traj_ids
        }
        return cls(trajs)

    @classmethod
    def from_csv(cls, fname: str, threshold=None, traj_id="id", time_id="time"):
        """
        Create TrajectoriesData DAO from CSV File
            To be a valid CSV the file must have a "time" field to record the sample time and an "id" field to distinguish
            the different trajectories.

        :param fname: CSV filename (path included).
        :param threshold: trajectories threshold value.
        :param traj_id: column name for trajectory identifier.
        :param time_id: column name for time identifier.

        :returns: TrajectoriesData

        Examples
            .. code-block:: python

                import autokoopman.core.trajectory as atraj

                trajs = atraj.TrajectoriesData.from_csv("./path/my_csv.csv")
        """
        data_df = pd.read_csv(fname)
        return cls.from_pandas(
            data_df, threshold=threshold, traj_id=traj_id, time_id=time_id
        )

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert to Pandas DataFrame

        Examples
            .. code-block:: python

                import autokoopman.benchmark.prde20 as prde

                sys = prde.ProdDestr()
                trajs = sys.solve_ivps(
                    initial_states=[[9.98, 0.01, 0.01]],
                    tspan=(0.0, 20.0),
                    sampling_period=0.01
                )
                trajs.to_pandas()
        """
        # form the dataframe
        serial_data = np.vstack(
            [
                np.hstack(
                    [
                        traji.times[:, np.newaxis],
                        traji.states,
                        np.array([namei] * len(traji.times), dtype=type(namei))[
                            :, np.newaxis
                        ],
                    ]
                )
                for idx, (namei, traji) in enumerate(self._trajs.items())
            ]
        )
        columns = [self.time_id_hname, *self.state_names, self.traj_id_hname]
        return pd.DataFrame(columns=columns, data=serial_data)

    def to_csv(self, fname: str):
        """
        Convert to CSV File (serialization)

        Examples
            .. code-block:: python

                import autokoopman.benchmark.prde20 as prde

                sys = prde.ProdDestr()
                trajs = sys.solve_ivps(
                    initial_states=[[9.98, 0.01, 0.01]],
                    tspan=(0.0, 20.0),
                    sampling_period=0.01
                )
                trajs.to_csv()
        """
        df = self.to_pandas()
        # write to csv
        df.to_csv(fname, index=False)

    def interp_uniform_time(self, sampling_period) -> "UniformTimeTrajectoriesData":
        return UniformTimeTrajectoriesData(
            {k: v.interp_uniform_time(sampling_period) for k, v in self._trajs.items()}
        )

    def __init__(self, trajs: Dict[Hashable, Union[UniformTimeTrajectory, Trajectory]]):
        self._trajs: Dict[Hashable, Union[UniformTimeTrajectory, Trajectory]] = trajs
        self._names_list = [v.names for _, v in self._trajs.items()]
        self._input_names_list = [v.input_names for _, v in self._trajs.items()]
        assert self.equal_lists(self._names_list), "all state names must be the same"
        assert self.equal_lists(
            self._input_names_list
        ), "all input names must be the same"
        self._state_names = self._names_list[0]
        self._input_names = self._input_names_list[0]

    def get_trajectory(self, traj_id: Hashable) -> Trajectory:
        return self._trajs[traj_id]

    @property
    def n_trajs(self) -> int:
        return len(self._trajs)

    @property
    def state_names(self) -> Sequence[str]:
        return self._state_names

    @property
    def input_names(self) -> Optional[Sequence[str]]:
        return self._input_names

    @property
    def traj_names(self) -> Set[Hashable]:
        return set(self._trajs.keys())

    def __getitem__(self, item):
        return self.get_trajectory(item)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} N Trajs: {self.n_trajs}, Traj "
            f"Names: [{_clip_list(list(self.traj_names))}], State "
            f"Names: [{_clip_list(self.state_names)}], Input "
            f"Names: [{_clip_list(self.input_names)}]>"
        )

    def __iter__(self):
        for _, v in self._trajs.items():
            yield v

    def __len__(self):
        return len(self._trajs)

    def _prepare_other(self, other: "TrajectoriesData"):
        assert set(other.traj_names) == set(
            self.traj_names
        ), "other data traj names must have same names"
        return other

    def __sub__(self, other: "TrajectoriesData"):
        otheri = self._prepare_other(other)
        return TrajectoriesData({k: v - otheri[k] for k, v in self._trajs.items()})

    def __add__(self, other: "TrajectoriesData"):
        otheri = self._prepare_other(other)
        return TrajectoriesData({k: v + otheri[k] for k, v in self._trajs.items()})

    def norm(self):
        return TrajectoriesData({k: v.norm() for k, v in self._trajs.items()})


class UniformTimeTrajectoriesData(TrajectoriesData):
    """a dataset of uniform time trajectories"""

    def __init__(self, trajs: Dict[Hashable, Union[UniformTimeTrajectory, Trajectory]]):
        super(UniformTimeTrajectoriesData, self).__init__(trajs)

    @property
    def sampling_period(self) -> float:
        """
        Sampling Period
            A sampling period allows the trajectories to relate to discrete and continuous time signals. Given a
            discrete time signal :math:`x_d`, it relates to an analog signal :math:`x_a` by a sampling time
            :math:`T_s`,

            .. math::

                x_a(n T_s) = x_d(n)
        """
        return [
            v.sampling_period
            for _, v in self._trajs.items()
            if isinstance(v, UniformTimeTrajectory)
        ][0]

    @classmethod
    def from_pandas(
        cls, data_df: pd.DataFrame, threshold=None, time_id="time", traj_id="id"
    ):
        """
        Create UniformTimeTrajectoriesData DAO from Pandas DataFrame
            To be a valid DataFrame the object must have a "time" field to record the sample time and an "id" field to distinguish
            the different trajectories.

        :param data_df: Pandas Dataframe to convert to trajectories.
        :param threshold: trajectories threshold value.
        :param traj_id: column name for trajectory identifier.
        :param time_id: column name for time identifier.

        :returns: TrajectoriesData

        Examples
            .. code-block:: python

                import autokoopman.core.trajectory as atraj
                import pandas as pd

                # create example dataframe
                t = np.linspace(0, 1.0, 10)
                my_df = pd.DataFrame(
                    columns=["time", "state", "id"],
                    data=np.array([
                        t,
                        np.sin(t),
                        np.ones(shape=t.shape)
                    ]).T
                )

                # create UniformTimeTrajectoriesData from the DataFrame
                trajs = atraj.UniformTimeTrajectoriesData.from_pandas(my_df)
        """
        traj_data = super().from_pandas(
            data_df, threshold=threshold, time_id=time_id, traj_id=traj_id
        )
        for k, v in traj_data._trajs.items():
            assert v.is_uniform_time, f"{k} is not uniform time"
        return cls({k: v.to_uniform_time_traj() for k, v in traj_data._trajs.items()})

    @property
    def next_step_matrices(
        self, nstep=1
    ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
        r"""
        Next Step Snapshot Matrices
            Return the two "snapshot matrices" :math:`\mathbf X, \mathbf X'` of observations :math:`\{x_1, x_2, ..., x_n \}`,

            .. math::

                \mathbf X = \begin{bmatrix} x_1 && x_2 && ... && x_{n-1}  \end{bmatrix}, \quad \mathbf X = \begin{bmatrix} x_2 && x_3 && ... && x_{n}  \end{bmatrix}

            Note this is useful for DMD techniques where the relationship

            .. math::

                \mathbf X' \approx A \mathbf X

            is used. Importantly, note that the observations are column vectors, which is counter to convention commonly
            used in vector processing.

        :param nstep: number of time steps in the future

        :returns: tuple of (:math:`\mathbf X, \mathbf X'`).

        References
            Brunton, S. L., & Kutz, J. N. (2022). Data-driven science and engineering: Machine learning,
            dynamical systems, and control. Cambridge University Press. pp 236-7
        """
        return self.n_step_matrices(1)

    def n_step_matrices(
        self, nstep
    ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
        X = np.vstack([x.states[:-nstep:nstep, :] for _, x in self._trajs.items()]).T

        Xp = np.vstack([x.states[nstep::nstep, :] for _, x in self._trajs.items()]).T

        if self.input_names is not None:
            U = np.vstack(
                [u.inputs[:-nstep:nstep, :] for _, u in self._trajs.items()]
            ).T
        else:
            U = None

        return X, Xp, U

    @property
    def differentiate(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""
        Numerical Differentiation -- Current State, Gradient State Estimate, Input
        """
        X = np.vstack([x.states[:-1, :] for _, x in self._trajs.items()]).T

        # finite difference
        Xp = np.vstack([x.states[1:, :] for _, x in self._trajs.items()]).T
        Xp -= X
        Xp /= self.sampling_period

        if self.input_names is not None:
            U = np.vstack([u.inputs[:-1, :] for _, u in self._trajs.items()]).T
        else:
            U = None

        return X, Xp, U



import numpy as np

from autokoopman.core.trajectory import TrajectoriesData
from autokoopman.core.tuner import (
    TuneResults,
    TrajectoryScoring,
    HyperparameterMap,
    ParameterSpace,
)
import autokoopman.core.tuner as atuner
import itertools
from typing import Callable


class GridSearchTuner(atuner.HyperparameterTuner):
    @staticmethod
    def make_grid(space: ParameterSpace, n_samps):
        parameters = []
        for coord in space:
            if isinstance(coord, atuner.ContinuousParameter):
                if coord.distribution == "loguniform":
                    elems = np.logspace(
                        np.log10(coord._interval[0]),
                        np.log10(coord._interval[1]),
                        num=n_samps,
                    )
                    parameters.append(elems)
                elif coord.distribution:
                    parameters.append(
                        np.linspace(coord._interval[0], coord._interval[1], num=n_samps)
                    )
            elif isinstance(coord, atuner.FiniteParameter):
                parameters.append(list(coord.elements))
        return parameters

    def __init__(
        self,
        parameter_model: HyperparameterMap,
        training_data: TrajectoriesData,
        n_samps=10,
        **kwargs,
    ):
        super(GridSearchTuner, self).__init__(parameter_model, training_data, **kwargs)
        self.n_samps = n_samps

    def tune(
        self,
        nattempts=100,
        scoring_func: Callable[
            [TrajectoriesData, TrajectoriesData], float
        ] = TrajectoryScoring.end_point_score,
    ) -> TuneResults:
        # get the number of samples created from the meshgrid
        parameters = self.make_grid(self._parameter_model.parameter_space, self.n_samps)
        nsamples = np.prod(np.array([len(p) for p in parameters]))

        # create the sampler and send it the parameters
        sampling = self.tune_sampling(nsamples, scoring_func)
        next(sampling)

        for param in itertools.product(*parameters):
            try:
                sampling.send(param)
                next(sampling)
            except StopIteration:
                break
            except Exception as exc:
                print(f"Error: {exc}")
                self.error_messages.append((param, exc))
        return self.best_result



import autokoopman.benchmark.fhn as afhn
import numpy as np


def test_fhn():
    fhn = afhn.FitzHughNagumo()
    training_ivs = np.random.random((10, 2))
    fhn.solve_ivps(training_ivs, (0.0, 3.0), sampling_period=0.05)
